import typing as t
from django.conf import settings
from drf_spectacular.utils import extend_schema_field, extend_schema_serializer
from drf_spectacular.types import OpenApiTypes
from rest_framework.reverse import reverse
from rest_framework import serializers

from pulpcore.plugin.util import get_domain
from pulpcore.plugin.models import RepositoryVersion
from pulpcore.plugin import serializers as core_serializers

from pulp_ansible.app import models, fields, serializers as ansible_serializers

DOMAIN_ENABLED = settings.DOMAIN_ENABLED


def _get_distro_context(context):
    distro_context = {}
    if path := context.get("path") or context.get("distro_base_path"):
        distro_context["path"] = path
    if "distro_base_path" in context:
        distro_context["distro_base_path"] = context["distro_base_path"]
    if DOMAIN_ENABLED:
        distro_context["pulp_domain"] = get_domain().name
    return distro_context


class CollectionSerializer(serializers.ModelSerializer):
    """A serializer for a Collection."""

    deprecated = serializers.BooleanField()
    created_at = serializers.SerializerMethodField()
    updated_at = serializers.SerializerMethodField()
    href = serializers.SerializerMethodField()

    versions_url = serializers.SerializerMethodField()
    highest_version = serializers.SerializerMethodField()
    download_count = serializers.SerializerMethodField()

    class Meta:
        fields = (
            "href",
            "namespace",
            "name",
            "deprecated",
            "versions_url",
            "highest_version",
            "created_at",
            "updated_at",
            "download_count",
        )
        model = models.Collection

    def get_href(self, obj) -> str:
        """Get href."""
        ctx = _get_distro_context(self.context)
        return reverse(
            settings.ANSIBLE_URL_NAMESPACE + "collections-detail",
            kwargs={**ctx, "namespace": obj.namespace, "name": obj.name},
        )

    def get_versions_url(self, obj) -> str:
        """Get a link to a collection versions list."""
        ctx = _get_distro_context(self.context)
        return reverse(
            settings.ANSIBLE_URL_NAMESPACE + "collection-versions-list",
            kwargs={**ctx, "namespace": obj.namespace, "name": obj.name},
        )

    @extend_schema_field(OpenApiTypes.DATETIME)
    def get_created_at(self, obj):
        """Get the timestamp of the lowest version CollectionVersion's created timestamp."""
        return obj.pulp_created

    @extend_schema_field(OpenApiTypes.DATETIME)
    def get_updated_at(self, obj):
        """Get the timestamp of the latest version CollectionVersion's created timestamp."""
        # NOTE: this should reflect the last time the collection was updated, not the last
        # time that the latest version was updated. See https://pulp.plan.io/issues/7775
        return obj.latest_version_modified

    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_highest_version(self, obj):
        """Get a highest version and its link."""
        ctx = _get_distro_context(self.context)
        href = reverse(
            settings.ANSIBLE_URL_NAMESPACE + "collection-versions-detail",
            kwargs={
                **ctx,
                "namespace": obj.namespace,
                "name": obj.name,
                "version": obj.highest_version,
            },
        )
        return {"href": href, "version": obj.highest_version}

    def get_download_count(self, obj) -> int:
        """Get the download count of the collection"""

        return obj.download_count or 0


class CollectionVersionListSerializer(serializers.ModelSerializer):
    """A serializer for a CollectionVersion list item."""

    href = serializers.SerializerMethodField()
    created_at = serializers.DateTimeField(source="pulp_created")
    updated_at = serializers.DateTimeField(source="pulp_last_updated")
    marks = serializers.SerializerMethodField()

    class Meta:
        fields = (
            "version",
            "href",
            "created_at",
            "updated_at",
            "requires_ansible",
            "marks",
        )
        model = models.CollectionVersion

    def get_marks(self, obj) -> t.List[str]:
        """Get a list of mark values filtering only those in the current repo."""
        return [x.value for x in obj.marks.all()]

    def get_href(self, obj) -> str:
        """
        Get href.
        """
        ctx = _get_distro_context(self.context)

        return reverse(
            settings.ANSIBLE_URL_NAMESPACE + "collection-versions-detail",
            kwargs={
                **ctx,
                "namespace": obj.namespace,
                "name": obj.name,
                "version": obj.version,
            },
        )


class ArtifactRefSerializer(serializers.Serializer):
    """A serializer for an Artifact reference."""

    filename = serializers.CharField(source="relative_path")
    sha256 = serializers.CharField(source="artifact.sha256")
    size = serializers.IntegerField(source="artifact.size")


class CollectionRefSerializer(serializers.Serializer):
    """
    A serializer for a Collection reference.
    """

    id = serializers.CharField(source="pk")
    name = serializers.CharField()
    href = serializers.SerializerMethodField()

    def get_href(self, obj) -> str:
        """Returns link to a collection."""
        ctx = _get_distro_context(self.context)
        return reverse(
            settings.ANSIBLE_URL_NAMESPACE + "collections-detail",
            kwargs={**ctx, "namespace": obj.namespace, "name": obj.name},
        )


class CollectionMetadataSerializer(serializers.ModelSerializer):
    """
    A serializer for a CollectionVersion metadata.
    """

    class Meta:
        model = models.CollectionVersion
        fields = (
            "authors",
            "contents",
            "dependencies",
            "description",
            "documentation",
            "homepage",
            "issues",
            "license",
            "repository",
            "tags",
        )


class CollectionNamespaceSerializer(serializers.Serializer):
    """
    A serializer for a Collection Version namespace field.
    """

    name = serializers.CharField(source="namespace")
    metadata_sha256 = serializers.CharField(source="namespace_sha256", allow_null=True)


class CollectionVersionSignatureSerializer(serializers.ModelSerializer):
    """
    A serializer for the signatures on a Collection Version.
    """

    signature = serializers.SerializerMethodField()
    signing_service = serializers.SlugRelatedField(
        slug_field="name",
        allow_null=True,
        read_only=True,
    )

    def get_signature(self, obj):
        """
        Get the signature data.
        """
        return obj.data

    class Meta:
        model = models.CollectionVersionSignature
        fields = ("signature", "pubkey_fingerprint", "signing_service", "pulp_created")


class UnpaginatedCollectionVersionSerializer(CollectionVersionListSerializer):
    """
    A serializer for unpaginated CollectionVersion.
    """

    collection = CollectionRefSerializer(read_only=True)
    artifact = serializers.SerializerMethodField()
    download_url = serializers.SerializerMethodField()
    git_url = serializers.SerializerMethodField()
    git_commit_sha = serializers.SerializerMethodField()

    metadata = CollectionMetadataSerializer(source="*", read_only=True)
    namespace = CollectionNamespaceSerializer(source="*", read_only=True)
    signatures = CollectionVersionSignatureSerializer(many=True)

    class Meta:
        model = models.CollectionVersion
        fields = CollectionVersionListSerializer.Meta.fields + (
            "artifact",
            "collection",
            "download_url",
            "name",
            "namespace",
            "signatures",
            "metadata",
            "git_url",
            "git_commit_sha",
        )

    @extend_schema_field(ArtifactRefSerializer)
    def get_artifact(self, obj):
        """
        Get artifact summary.
        """
        ca = obj.artifacts[0]
        if ca.artifact:
            return ArtifactRefSerializer(ca).data

    def get_download_url(self, obj) -> str:
        """
        Get artifact download URL.
        """
        if obj.artifacts[0].artifact:
            distro_base_path = self.context.get("path", self.context["distro_base_path"])
            filename_path = obj.relative_path.lstrip("/")

            # Note: We're using ANSIBLE_API_HOSTNAME here instead of calling reverse with request=
            # because using the request context to generate the full URL causes the download URL
            # to be inaccessible when pulp is running behind a reverse proxy.
            host = settings.ANSIBLE_API_HOSTNAME.strip("/")
            extra_args = {}
            if DOMAIN_ENABLED:
                extra_args["pulp_domain"] = get_domain().name
            path = reverse(
                settings.ANSIBLE_URL_NAMESPACE + "collection-artifact-download",
                kwargs={
                    "distro_base_path": distro_base_path,
                    "filename": filename_path,
                    **extra_args,
                },
            ).strip("/")

            return f"{host}/{path}"

    def get_git_url(self, obj) -> str:
        """
        Get the git URL.
        """
        ca = obj.artifacts[0]
        if not ca.artifact:
            return ca.remoteartifact_set.all()[0].url[:-47]

    def get_git_commit_sha(self, obj) -> str:
        """
        Get the git commit sha.
        """
        ca = obj.artifacts[0]
        if not ca.artifact:
            return ca.remoteartifact_set.all()[0].url[-40:]


class CollectionVersionSerializer(UnpaginatedCollectionVersionSerializer):
    """
    A serializer for a CollectionVersion.
    """

    manifest = fields.JSONDictField(
        help_text="A JSON field holding MANIFEST.json data.", read_only=True
    )
    files = fields.JSONDictField(help_text="A JSON field holding FILES.json data.", read_only=True)

    class Meta:
        model = models.CollectionVersion
        fields = UnpaginatedCollectionVersionSerializer.Meta.fields + (
            "manifest",
            "files",
        )


class CollectionVersionDocsSerializer(serializers.ModelSerializer):
    """A serializer to display the docs_blob of a CollectionVersion."""

    docs_blob = fields.JSONDictField()

    class Meta:
        fields = ("docs_blob",)
        model = models.CollectionVersion


class RepoMetadataSerializer(serializers.ModelSerializer):
    """A serializer to display RepositoryVersion metadata."""

    published = serializers.DateTimeField(source="pulp_created")

    class Meta:
        fields = ("published",)
        model = RepositoryVersion


class ClientConfigurationSerializer(serializers.Serializer):
    """Configuration settings for the ansible-galaxy client."""

    default_distribution_path = serializers.CharField(allow_null=True)


@extend_schema_serializer(component_name="CollectionSummary")
class CollectionSummarySerializer(ansible_serializers.CollectionVersionSerializer):
    """Collection Version serializer without docs blob."""

    class Meta:
        model = models.CollectionVersion
        fields = (
            "pulp_href",
            "namespace",
            "name",
            "version",
            "requires_ansible",
            "pulp_created",
            "contents",
            "dependencies",
            "description",
            "tags",
        )


class NamespaceSummarySerializer(ansible_serializers.AnsibleNamespaceMetadataSerializer):
    """Namespace serializer without resources."""

    class Meta:
        model = models.AnsibleNamespaceMetadata
        fields = (
            "pulp_href",
            "name",
            "company",
            "description",
            "avatar",
            "avatar_url",
        )


class CollectionVersionSearchListSerializer(CollectionVersionListSerializer):
    """Cross-repo search results."""

    # All of these fields have to operate differently from the parent class
    repository = core_serializers.RepositorySerializer()
    collection_version = CollectionSummarySerializer()
    repository_version = serializers.SerializerMethodField()
    namespace_metadata = NamespaceSummarySerializer(allow_null=True)

    is_highest = serializers.BooleanField()
    is_signed = serializers.BooleanField()
    is_deprecated = serializers.BooleanField()

    class Meta:
        model = models.CrossRepositoryCollectionVersionIndex

        fields = (
            "repository",
            "collection_version",
            "repository_version",
            "namespace_metadata",
            "is_highest",
            "is_deprecated",
            "is_signed",
        )
        # This is a read only serializer.
        # But the uniqueness of the underlying model interfers with the SerilizerMethodField.
        # See https://github.com/encode/django-rest-framework/pull/9531#issuecomment-2778892573 .
        # This is a workaround. I don't even expect it to work forever.
        validators = []

    def get_repository_version(self, obj) -> str:
        if obj.repository_version:
            return str(obj.repository_version.number)
        else:
            return "latest"



# Generated by Django 2.2.13 on 2020-06-09 21:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0018_fix_collection_relative_path'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionremote',
            name='auth_url',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='collectionremote',
            name='token',
            field=models.TextField(max_length=2000, null=True),
        ),
    ]



import itertools
import random
import string

from django.db import connection
from django.test import TestCase

from pulp_ansible.app.models import (
    Collection,
    CollectionVersion,
)


class TestSearchUtil(TestCase):
    collections = None

    def setUp(self):
        # define 12 random collection version specifications
        self.collections = {}
        for i in range(0, 2):
            namespace_name = "".join([random.choice(string.ascii_lowercase) for x in range(0, 5)])
            for j in range(0, 3):
                collection_name = "".join(
                    [random.choice(string.ascii_lowercase) for x in range(0, 5)]
                )
                for v in range(1, 3):
                    vstring = f"1.0.{v}"
                    self.collections[(namespace_name, collection_name, vstring)] = {
                        "tags": ["tag" + namespace_name + collection_name + vstring],
                    }

        # we want to iterate in a sorted order
        specs = sorted(list(self.collections.keys()))

        # "import" the collections
        for ids, spec in enumerate(specs):
            cdata = self.collections[spec]

            # make the collection
            col, _ = Collection.objects.get_or_create(name=spec[0])
            col.save()
            self.collections[spec]["col"] = col

            # make the collection version
            cv = CollectionVersion.objects.create(
                collection=col,
                sha256=str(ids),
                namespace=spec[0],
                name=spec[1],
                version=spec[2],
                tags=cdata["tags"],
            )
            self.collections[spec]["cv"] = cv

            # trigger an update and rebuild of the search vector
            #   THIS IS THE ONLY WAY THAT THE SEARCH VECTOR IS CREATED!
            cv.is_highest = False
            cv.save()

    def tearDown(self):
        # delete collectionversions
        for spec, cdata in self.collections.items():
            cdata["cv"].delete()

        # delete the collections
        collections = [x["col"] for x in self.collections.values()]
        collections_deleted = []
        for collection in collections:
            if collection.name in collections_deleted:
                continue
            collection.delete()
            collections_deleted.append(collection.name)

    def test_search_vector_has_correct_tags(self):
        # flatten a list of all tags
        all_tags = list(itertools.chain(*[x["tags"] for x in self.collections.values()]))

        # fetch the search vector for every CV
        cursor = connection.cursor()
        cursor.execute("select namespace,name,version,search_vector from ansible_collectionversion")
        rows = cursor.fetchall()

        # make a map for quicker reference
        rmap = {}
        for row in rows:
            rmap[(row[0], row[1], row[2])] = row[3]

        # check the search vector on each collectionversion ...
        for ckey, cdata in self.collections.items():
            search_vector = rmap[ckey]
            assert search_vector != "", "search vector was not built"

            # ensure the right tags got in the vector
            good_matches = [x for x in cdata["tags"] if x in search_vector]
            assert sorted(good_matches) == sorted(cdata["tags"])

            # ensure no other tags got in the vector
            bad_tags = [x for x in all_tags if x not in cdata["tags"]]
            bad_matches = [x for x in bad_tags if x in search_vector]
            assert not bad_matches, f"found unrelated tags in {ckey} search vector {search_vector}"



from django.core.files.uploadedfile import SimpleUploadedFile


def test_collection_version_sha256_migrate(migrate):
    apps = migrate([("ansible", "0055_alter_collectionversion_version_alter_role_version")])
    Artifact = apps.get_model("core", "Artifact")
    Collection = apps.get_model("ansible", "Collection")
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")

    collection = Collection.objects.create(namespace="snap", name="crackle")

    # Create two collection versions, because `sha256=null` can violate the uniquenes constraint.
    artifact1 = Artifact.objects.create(
        size=8, sha256="SENTINEL1", file=SimpleUploadedFile("foo", b"deadbeef")
    )
    cv1 = CollectionVersion.objects.create(
        pulp_type="collection_version",
        collection=collection,
        namespace="snap",
        name="crackle",
        version="1.0.0",
        version_minor="1",
        version_major="0",
        version_patch="0",
    )
    cv1.contentartifact_set.create(artifact=artifact1)

    artifact2 = Artifact.objects.create(
        size=8, sha256="SENTINEL2", file=SimpleUploadedFile("foo", b"beefdead")
    )
    cv2 = CollectionVersion.objects.create(
        pulp_type="collection_version",
        collection=collection,
        namespace="snap",
        name="crackle",
        version="2.0.0",
        version_minor="2",
        version_major="0",
        version_patch="0",
    )
    cv2.contentartifact_set.create(artifact=artifact2)

    apps = migrate([("ansible", "0056_collectionversion_sha256")])
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")

    cv1 = CollectionVersion.objects.get(pk=cv1.pk)
    assert cv1.sha256 == ""

    apps = migrate([("ansible", "0057_collectionversion_sha256_migrate")])
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")

    cv1 = CollectionVersion.objects.get(pk=cv1.pk)
    assert cv1.sha256 == "SENTINEL1"

    apps = migrate([("ansible", "0059_collectionversion_unique_sha256")])
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")

    cv1 = CollectionVersion.objects.get(pk=cv1.pk)
    assert cv1.sha256 == "SENTINEL1"



# Generated by Django 2.2.15 on 2020-08-27 21:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0021_rename_role_remote'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='documentation',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='homepage',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='issues',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='repository',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
    ]



import uuid
import pytest
import subprocess
import time
import yaml
from types import SimpleNamespace

from pulpcore.tests.functional.utils import BindingsNamespace
from pulp_ansible.tests.functional.constants import ANSIBLE_FIXTURE_URL
from pulp_ansible.tests.functional.utils import randstr


# Bindings API Fixtures


@pytest.fixture(scope="session")
def ansible_bindings(_api_client_set, bindings_cfg):
    """
    A namespace providing preconfigured pulp_ansible api clients.

    e.g. `ansible_bindings.RepositoriesAnsibleApi.list()`.
    """
    from pulpcore.client import pulp_ansible as bindings_module

    api_client = bindings_module.ApiClient(bindings_cfg)
    _api_client_set.add(api_client)
    yield BindingsNamespace(bindings_module, api_client)
    _api_client_set.remove(api_client)


# Object Generation Fixtures


@pytest.fixture
def ansible_repo(ansible_repository_factory):
    """Creates an Ansible Repository and deletes it at test cleanup time."""
    return ansible_repository_factory()


@pytest.fixture(scope="class")
def ansible_repository_factory(ansible_bindings, gen_object_with_cleanup):
    """A factory that creates an Ansible Repository and deletes it at test cleanup time."""

    def _ansible_repository_factory(**kwargs):
        extra_args = {}
        if pulp_domain := kwargs.pop("pulp_domain", None):
            extra_args["pulp_domain"] = pulp_domain
        kwargs.setdefault("name", str(uuid.uuid4()))
        return gen_object_with_cleanup(
            ansible_bindings.RepositoriesAnsibleApi, kwargs, **extra_args
        )

    return _ansible_repository_factory


ansible_repo_factory = ansible_repository_factory


@pytest.fixture(scope="class")
def ansible_sync_factory(ansible_bindings, ansible_repo_factory, monitor_task):
    """A factory to perform a sync on an Ansible Repository and return its updated data."""

    def _sync(ansible_repo=None, **kwargs):
        body = ansible_bindings.module.AnsibleRepositorySyncURL(**kwargs)
        if ansible_repo is None:
            ansible_repo = ansible_repo_factory()
        monitor_task(
            ansible_bindings.RepositoriesAnsibleApi.sync(ansible_repo.pulp_href, body).task
        )
        return ansible_bindings.RepositoriesAnsibleApi.read(ansible_repo.pulp_href)

    return _sync


@pytest.fixture(scope="class")
def ansible_distribution_factory(ansible_bindings, gen_object_with_cleanup):
    """A factory to generate an Ansible Distribution with auto-cleanup."""

    def _ansible_distribution_factory(repository=None, **kwargs):
        extra_args = {}
        if pulp_domain := kwargs.pop("pulp_domain", None):
            extra_args["pulp_domain"] = pulp_domain
        kwargs.setdefault("name", str(uuid.uuid4()))
        kwargs.setdefault("base_path", str(uuid.uuid4()))
        if repository:
            kwargs["repository"] = repository.pulp_href
        return gen_object_with_cleanup(
            ansible_bindings.DistributionsAnsibleApi, kwargs, **extra_args
        )

    return _ansible_distribution_factory


@pytest.fixture(scope="class")
def ansible_role_remote_factory(bindings_cfg, ansible_bindings, gen_object_with_cleanup):
    """A factory to generate an Ansible Collection Remote with auto-cleanup."""

    def _ansible_role_remote_factory(include_pulp_auth=False, **kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        kwargs.setdefault("url", ANSIBLE_FIXTURE_URL)
        if include_pulp_auth:
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        return gen_object_with_cleanup(ansible_bindings.RemotesRoleApi, kwargs)

    return _ansible_role_remote_factory


@pytest.fixture(scope="class")
def ansible_collection_remote_factory(bindings_cfg, ansible_bindings, gen_object_with_cleanup):
    """A factory to generate an Ansible Collection Remote with auto-cleanup."""

    def _ansible_collection_remote_factory(include_pulp_auth=False, **kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        if include_pulp_auth:
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        # Ratelimiting, because our tests hit galaxy pretty hard.
        kwargs["rate_limit"] = 5

        return gen_object_with_cleanup(ansible_bindings.RemotesCollectionApi, kwargs)

    return _ansible_collection_remote_factory


@pytest.fixture(scope="class")
def ansible_git_remote_factory(bindings_cfg, ansible_bindings, gen_object_with_cleanup):
    """A factory to generate an Ansible Git Remote with auto-cleanup."""

    def _ansible_git_remote_factory(include_pulp_auth=False, **kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        if include_pulp_auth:
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        return gen_object_with_cleanup(ansible_bindings.RemotesGitApi, kwargs)

    return _ansible_git_remote_factory


@pytest.fixture(scope="session")
def ansible_collection_factory(tmp_path_factory):
    def _collection_factory(config=None):
        config = {} if config is None else config.copy()
        tmpdir = tmp_path_factory.mktemp("collection")
        namespace = config.setdefault("namespace", randstr())
        name = config.setdefault("name", randstr())
        config.setdefault("version", "1.0.0")

        src_path = tmpdir / namespace / name

        # Template the collection
        subprocess.run(f"ansible-galaxy collection init {namespace}.{name}", shell=True, cwd=tmpdir)

        # Adjust the template
        (src_path / "meta" / "runtime.yml").write_text('requires_ansible: ">=2.13"\n')
        (src_path / "README.md").write_text("# title\ncollection docs\n")
        galaxy_yaml = yaml.safe_load((src_path / "galaxy.yml").read_text())
        galaxy_yaml.update(config)
        (src_path / "galaxy.yml").write_text(yaml.safe_dump(galaxy_yaml))

        # Build the collection artifact
        build_proc = subprocess.run(
            "ansible-galaxy collection build .",
            shell=True,
            cwd=src_path,
            stdout=subprocess.PIPE,
            check=True,
        )
        filename = build_proc.stdout.decode("utf-8").strip().split()[-1]

        return SimpleNamespace(filename=filename, **config)

    return _collection_factory


@pytest.fixture(scope="class")
def build_and_upload_collection(ansible_bindings, monitor_task, ansible_collection_factory):
    """A factory to locally create, build, and upload a collection."""

    def _build_and_upload_collection(ansible_repo=None, **kwargs):
        collection = ansible_collection_factory(**kwargs)
        body = {"file": collection.filename}
        if ansible_repo:
            body["repository"] = ansible_repo.pulp_href
        response = ansible_bindings.ContentCollectionVersionsApi.create(**body)
        task = monitor_task(response.task)
        collection_href = [
            href for href in task.created_resources if "content/ansible/collection_versions" in href
        ]
        return collection, collection_href[0]

    return _build_and_upload_collection


@pytest.fixture
def ansible_dir_factory(tmp_path, bindings_cfg):
    """A factory to create a local ansible.cfg file with authentication"""

    def _ansible_dir_factory(server):
        username = bindings_cfg.username
        password = bindings_cfg.password

        ansible_cfg = (
            "[galaxy]\nserver_list = pulp_ansible\n\n"
            "[galaxy_server.pulp_ansible]\n"
            f"url = {server}\n"
        )

        if username is not None and password is not None:
            ansible_cfg += f"username = {username}\npassword = {password}\n"

        cfg_file = tmp_path / "ansible.cfg"
        cfg_file.write_text(ansible_cfg, encoding="UTF_8")
        return tmp_path

    return _ansible_dir_factory


@pytest.fixture(scope="session")
def random_image_factory(tmp_path_factory):
    """Factory to produce a random 100x100 image."""
    import numpy as np
    from PIL import Image

    def _random_image():
        imarray = np.random.rand(100, 100, 3) * 255
        im = Image.fromarray(imarray.astype("uint8")).convert("RGBA")
        path = tmp_path_factory.mktemp("images") / f"{randstr()}.png"
        im.save(path)
        return path

    return _random_image


# Utility fixtures


@pytest.fixture()
def skip_on_galaxy(pulp_versions):
    if "galaxy" in pulp_versions:
        pytest.skip("This test is not desigend to run on galaxy installations.")


@pytest.fixture(scope="session")
def wait_tasks(pulpcore_bindings):
    """Polls the Task API until all tasks for a resource are in a completed state."""

    def _wait_tasks(resource):
        pending_tasks = pulpcore_bindings.TasksApi.list(
            state__in=["running", "waiting"], reserved_resources=resource
        )
        while pending_tasks.count:
            time.sleep(1)
            pending_tasks = pulpcore_bindings.TasksApi.list(
                state__in=["running", "waiting"], reserved_resources=resource
            )

    return _wait_tasks



# Generated by Django 2.2.17 on 2021-01-25 22:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0027_tag_length'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='namespace',
            field=models.CharField(editable=False, max_length=64),
        ),
    ]



# Generated by Django 2.2.17 on 2020-12-09 20:22
import logging
import json
import tarfile
import yaml
from json import JSONDecodeError
from yaml.error import YAMLError
from django.db import migrations, models

from pulp_ansible.app.tasks.utils import get_file_obj_from_tarball

log = logging.getLogger(__name__)


def set_requires_ansible_and_manifest_and_files_json(apps, schema_editor):
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")
    for collection_version in CollectionVersion.objects.all():
        artifact = collection_version.contentartifact_set.get().artifact
        with artifact.file.open() as artifact_file, tarfile.open(
            fileobj=artifact_file, mode="r"
        ) as tar:
            runtime_metadata = get_file_obj_from_tarball(
                tar, "meta/runtime.yml", artifact.file.name, raise_exc=False
            )
            if runtime_metadata:
                try:
                    runtime_yaml = yaml.safe_load(runtime_metadata)
                except YAMLError:
                    log.warning(
                        "CollectionVersion: '{namespace}.{name}-{version}' - 'meta/runtime.yml' is invalid yaml.".format(
                            namespace=collection_version.namespace,
                            name=collection_version.name,
                            version=collection_version.version,
                        )
                    )
                else:
                    try:
                        collection_version.requires_ansible = runtime_yaml.get("requires_ansible")
                    except AttributeError:
                        log.warning(
                            "CollectionVersion: '{namespace}.{name}-{version}' - 'meta/runtime.yml' is missing key 'requires_ansible'.".format(
                                namespace=collection_version.namespace,
                                name=collection_version.name,
                                version=collection_version.version,
                            )
                        )

            manifest = get_file_obj_from_tarball(
                tar, "MANIFEST.json", artifact.file.name, raise_exc=False
            )
            if manifest:
                try:
                    collection_version.manifest = json.load(manifest)
                except JSONDecodeError:
                    log.warning(
                        "CollectionVersion: '{namespace}.{name}-{version}' - 'MANIFEST.json' is invalid json.".format(
                            namespace=collection_version.namespace,
                            name=collection_version.name,
                            version=collection_version.version,
                        )
                    )

            files = get_file_obj_from_tarball(
                tar, "FILES.json", artifact.file.name, raise_exc=False
            )
            if files:
                try:
                    collection_version.files = json.load(files)
                except JSONDecodeError:
                    log.warning(
                        "CollectionVersion: '{namespace}.{name}-{version}' - 'FILES.json' is invalid json.".format(
                            namespace=collection_version.namespace,
                            name=collection_version.name,
                            version=collection_version.version,
                        )
                    )
            collection_version.save()


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0029_manifest_and_files_json_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="collectionversion",
            name="requires_ansible",
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.RunPython(
            code=set_requires_ansible_and_manifest_and_files_json,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        ),
    ]



import hashlib
import json
import subprocess
import os

import pytest

from pulpcore.client.pulp_ansible.exceptions import ApiException
from pulp_ansible.tests.functional.utils import randstr


def busted_collection_build(basedir=None, namespace=None, name=None, version=None):
    """Make artifacts with namespaces and names that wouldn't normally be allowed."""

    def file_checksum(fp):
        with open(fp, "rb") as f:
            checksum = hashlib.sha256(f.read()).hexdigest()
        return checksum

    # make a content dir
    content_dir = os.path.join(basedir, "content")
    os.makedirs(content_dir)

    # make various files
    for dname in ["docs", "plugins", "roles"]:
        os.makedirs(os.path.join(content_dir, dname))
    with open(os.path.join(content_dir, "README.md"), "w") as f:
        f.write("")
    with open(os.path.join(content_dir, "plugins", "README.md"), "w") as f:
        f.write("")

    # make files.json
    fdata = {"files": [], "format": 1}
    for root, dirs, files in os.walk(content_dir):
        for dirname in dirs:
            dirpath = os.path.join(root, dirname)
            relative_dir = dirpath.replace(content_dir + "/", "")
            fdata["files"].append(
                {
                    "name": relative_dir,
                    "ftype": "dir",
                    "chksum_type": None,
                    "chksum_sha256": None,
                    "format": 1,
                }
            )
        for filen in files:
            filepath = os.path.join(root, filen)
            relative_file = filepath.replace(content_dir + "/", "")
            fdata["files"].append(
                {
                    "name": relative_file,
                    "ftype": "file",
                    "chksum_type": "sha256",
                    "chksum_sha256": file_checksum(filepath),
                    "format": 1,
                }
            )
    with open(os.path.join(content_dir, "FILES.json"), "w") as f:
        f.write(json.dumps(fdata, indent=2))

    # make manifest.json
    mdata = {
        "format": 1,
        "collection_info": {
            "namespace": namespace,
            "name": name,
            "version": version,
            "authors": ["your name <example@domain.com>"],
            "readme": "README.md",
            "tags": [],
            "description": "your collection description",
            "license": ["GPL-2.0-or-later"],
            "license_file": None,
            "dependencies": {},
            "repository": "http://example.com/repository",
            "homepage": "http://example.com",
            "issues": "http://example.com/issue/tracker",
        },
        "file_manifest_file": {
            "name": "FILES.json",
            "ftype": "file",
            "chksum_type": "sha256",
            "chksum_sha256": file_checksum(os.path.join(content_dir, "FILES.json")),
            "format": 1,
        },
    }
    with open(os.path.join(content_dir, "MANIFEST.json"), "w") as f:
        f.write(json.dumps(mdata, indent=2))

    # make the tarball
    cmd = "tar czvf artifact.tar.gz *"
    subprocess.run(cmd, shell=True, cwd=content_dir)
    return os.path.join(content_dir, "artifact.tar.gz")


@pytest.mark.parallel
def test_collection_named_collections(
    ansible_bindings,
    tmp_path,
    ansible_repo,
    ansible_collection_factory,
    ansible_distribution_factory,
    monitor_task,
):
    """A CV with a name of 'collections' could break url parsing."""
    # https://issues.redhat.com/browse/AAH-2158
    pulp_dist = ansible_distribution_factory(ansible_repo)

    spec = {
        "repository": "https://github.com/foo/bar",
        "namespace": randstr(),
        "name": "collections",
        "version": "1.0.0",
    }
    artifact = ansible_collection_factory(config=spec).filename

    body = {"file": artifact}
    body["repository"] = ansible_repo.pulp_href
    response = ansible_bindings.ContentCollectionVersionsApi.create(**body)
    monitor_task(response.task)

    # validate the collection shows up ...
    resp = ansible_bindings.PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi.list(
        pulp_dist.base_path
    )
    assert resp.meta.count == 1
    assert resp.data[0].namespace == spec["namespace"]
    assert resp.data[0].name == spec["name"]

    # validate we can get the direct path to the collection ...
    resp = ansible_bindings.PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi.read(
        pulp_dist.base_path, namespace=spec["namespace"], name=spec["name"]
    )
    assert resp.namespace == spec["namespace"]
    assert resp.name == spec["name"]


@pytest.mark.parallel
def test_collection_named_with_slashes(
    ansible_bindings,
    tmp_path,
    ansible_repo,
    ansible_distribution_factory,
):
    """A CV with a slash in the name could break url parsing."""
    # https://issues.redhat.com/browse/AAH-2158
    ansible_distribution_factory(ansible_repo)

    spec = {
        "repository": "https://github.com/foo/bar",
        "namespace": "collections/are/fun",
        "name": "index/foo/myname",
        "version": "1.0.0",
    }

    tarball = busted_collection_build(
        basedir=tmp_path,
        namespace=spec["namespace"],
        name=spec["name"],
        version=spec["version"],
    )

    body = {"file": tarball}
    body["repository"] = ansible_repo.pulp_href

    with pytest.raises(ApiException):
        ansible_bindings.ContentCollectionVersionsApi.create(**body)



# Generated by Django 3.2.15 on 2022-08-29 11:04

from django.db import migrations
import pulpcore.app.models.fields


def encrypt_remote_fields(apps, schema_editor):
    CollectionRemote = apps.get_model("ansible", "CollectionRemote")

    remotes_needing_update = []
    for remote in CollectionRemote.objects.order_by("pulp_id").iterator():
        remote._encrypted_token = remote.token
        remotes_needing_update.append(remote)

        if len(remotes_needing_update) > 100:
            CollectionRemote.objects.bulk_update(remotes_needing_update, ("_encrypted_token",))
            remotes_needing_update.clear()

    CollectionRemote.objects.bulk_update(remotes_needing_update, ("_encrypted_token",))


def unencrypt_remote_fields(apps, schema_editor):
    CollectionRemote = apps.get_model("ansible", "CollectionRemote")

    remotes_needing_update = []
    for remote in CollectionRemote.objects.order_by("pulp_id").iterator():
        remote.token = remote._encrypted_token
        remotes_needing_update.append(remote)

        if len(remotes_needing_update) > 100:
            CollectionRemote.objects.bulk_update(remotes_needing_update, ("token",))
            remotes_needing_update.clear()

    CollectionRemote.objects.bulk_update(remotes_needing_update, ("token",))


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0043_alter_collectionversionsignature_data'),
    ]

    operations = [
        # Add new fields to temporarily hold the encrypted values
        migrations.AddField(
            model_name="collectionremote",
            name="_encrypted_token",
            field=pulpcore.app.models.fields.EncryptedTextField(null=True),
        ),
        # Populate the new fields with encrypted values computed from the unencrypted fields
        migrations.RunPython(
            code=encrypt_remote_fields,
            reverse_code=unencrypt_remote_fields,
            elidable=True,
        ),
        # Remove the unencrypted columns
        migrations.RemoveField(
            model_name="collectionremote",
            name="token",
        ),
        # Replace the formerly-unencrypted columns with the new encrypted ones
        migrations.RenameField(
            model_name="collectionremote",
            old_name="_encrypted_token",
            new_name="token",
        ),
    ]



# Generated by Django 2.2.15 on 2020-08-10 19:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0019_collection_token'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='ansiblerepository',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': (('modify_ansible_repo_content', 'Can modify ansible repository content'),)},
        ),
    ]



# Generated by Django 3.2.18 on 2023-03-01 23:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0048_collectionversionmark'),
    ]

    operations = [
        migrations.AddField(
            model_name='ansiblerepository',
            name='private',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterModelOptions(
            name='ansibledistribution',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_ansibledistribution', 'Can manage roles on distributions')]},
        ),
        migrations.AlterModelOptions(
            name='ansiblerepository',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('rebuild_metadata_ansiblerepository', 'Can rebuild metadata on the repository'), ('repair_ansiblerepository', 'Can repair the repository'), ('sign_ansiblerepository', 'Can sign content on the repository'), ('sync_ansiblerepository', 'Can start a sync task on the repository'), ('manage_roles_ansiblerepository', 'Can manage roles on repositories'), ('modify_ansible_repo_content', 'Can modify repository content')]},
        ),
        migrations.AlterModelOptions(
            name='collectionremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': (('manage_roles_collectionremote', 'Can manage roles on collection remotes'),)},
        ),
        migrations.AlterModelOptions(
            name='gitremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_gitremote', 'Can manage roles on git remotes')]},
        ),
        migrations.AlterModelOptions(
            name='roleremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_roleremote', 'Can manage roles on role remotes')]},
        ),
    ]



# Generated by Django 4.2.16 on 2024-12-05 12:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0036_update_repository_content'),
    ]

    operations = [
        migrations.CreateModel(
            name='GitRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_gitremote', serialize=False, to='core.remote')),
                ('metadata_only', models.BooleanField(default=False)),
                ('git_ref', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.remote',),
        ),
    ]



import json
import logging
import tarfile

from django.db import transaction

from galaxy_importer.collection import import_collection
from pulpcore.plugin.models import Task
from pulpcore.plugin.util import get_url

from pulp_ansible.app.models import Collection, CollectionImport
from pulp_ansible.app.tasks.utils import CollectionFilename, get_file_obj_from_tarball

log = logging.getLogger(__name__)


def process_collection_artifact(artifact, namespace, name, version):
    """
    Helper method to extract a Collection's metadata.

    This is called from ``CollectionVersionUploadSerializer.deferred_validate()``.
    """
    # Avoid circular import
    from .collections import _get_backend_storage_url

    # Set up logging for CollectionImport object
    CollectionImport.objects.get_or_create(task_id=Task.current().pulp_id)
    user_facing_logger = logging.getLogger(
        "pulp_ansible.app.tasks.upload.process_collection_artifact"
    )

    artifact_url = get_url(artifact)
    filename = CollectionFilename(namespace, name, version)
    log.info(f"Processing collection {filename} from {artifact_url}")

    # Extra CollectionVersion metadata
    with artifact.file.open() as artifact_file:
        url = _get_backend_storage_url(artifact_file)
        importer_result = import_collection(
            artifact_file, filename=filename, file_url=url, logger=user_facing_logger
        )
        artifact_file.seek(0)
        with tarfile.open(fileobj=artifact_file, mode="r") as tar:
            manifest_data = json.load(
                get_file_obj_from_tarball(tar, "MANIFEST.json", artifact.file.name)
            )
            files_data = json.load(get_file_obj_from_tarball(tar, "FILES.json", artifact.file.name))

    # Set CollectionVersion metadata
    collection_info = importer_result["metadata"]
    collection_info["sha256"] = artifact.sha256

    with transaction.atomic():
        collection, created = Collection.objects.get_or_create(
            namespace=collection_info["namespace"], name=collection_info["name"]
        )
    collection_info["collection"] = collection
    collection_info["manifest"] = manifest_data
    collection_info["files"] = files_data
    collection_info["requires_ansible"] = importer_result.get("requires_ansible")
    collection_info["contents"] = importer_result["contents"]
    collection_info["docs_blob"] = importer_result["docs_blob"]
    # Remove fields not used by this model
    collection_info.pop("license_file")
    collection_info.pop("readme")
    # the importer returns many None values. We need to let the defaults in the model prevail
    for key in ["description", "documentation", "homepage", "issues", "repository"]:
        if collection_info[key] is None:
            collection_info.pop(key)

    collection_info["relative_path"] = (
        f"{collection_info['namespace']}-{collection_info['name']}-{collection_info['version']}"
        ".tar.gz"
    )
    return collection_info



import pytest


@pytest.fixture(scope="class")
def seeded_distribution(
    ansible_collection_remote_factory,
    ansible_sync_factory,
    ansible_distribution_factory,
    monitor_task,
):
    """Set up the content guard tests."""
    requirements_file = (
        "collections:\n" "  - name: testing.k8s_demo_collection\n" "  - name: pulp.squeezer"
    )
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file=requirements_file,
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote.pulp_href)
    distribution = ansible_distribution_factory(repository=repository)
    return distribution


@pytest.fixture(scope="class")
def verify_list(ansible_bindings, seeded_distribution):
    def _verify_list(new_api, old_api, query_params, args=[]):
        def is_list_identical(r1, r2):
            d1 = r1.to_dict()
            d2 = r2.to_dict()

            # skip comparing links so that the tests pass when galaxy_ng is installed
            assert d1["meta"] == d2["meta"]
            assert d1["data"] == d2["data"]

        args_new = [seeded_distribution.base_path, *args]

        # test pagination
        r1 = new_api.list(*args_new, limit=1)
        r2 = old_api.list(*args, limit=1)

        assert len(r1.data) == 1
        assert len(r2.data) == 1

        is_list_identical(r1, r2)

        # test query params
        r1 = new_api.list(*args_new, **query_params)
        r2 = old_api.list(*args, **query_params)

        is_list_identical(r1, r2)

        # test no query params
        r1 = new_api.list(*args_new)
        r2 = old_api.list(*args)

        is_list_identical(r1, r2)

    return _verify_list


@pytest.fixture(scope="class")
def verify_all_collections(seeded_distribution):
    def _verify_all_collections(new_api, old_api):
        r1 = new_api.list(seeded_distribution.base_path, seeded_distribution.base_path)
        r2 = old_api.list(seeded_distribution.base_path)

        assert r1 == r2

    return _verify_all_collections


@pytest.fixture(scope="class")
def verify_read(ansible_bindings, seeded_distribution):
    def _verify_read(new_api, old_api, version=None):
        args = ["k8s_demo_collection", "testing", seeded_distribution.base_path]
        if version:
            args.append(version)

        r1 = new_api.read(seeded_distribution.base_path, *args)
        r2 = old_api.read(*args)

        assert r1 == r2

    return _verify_read


class TestRedirectsAreIdentical:
    """
    Verify new and legacy endpoints are identical.
    """

    def test_redirect_collections_list(self, ansible_bindings, seeded_distribution, verify_list):
        verify_list(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexApi,
            ansible_bindings.PulpAnsibleApiV3CollectionsApi,
            {"namespace": "testing"},
            args=[seeded_distribution.base_path],
        )

    def test_redirect_collections_versions_list(
        self, ansible_bindings, seeded_distribution, verify_list
    ):

        verify_list(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexVersionsApi,
            ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi,
            {},
            args=["k8s_demo_collection", "testing", seeded_distribution.base_path],
        )

    def test_redirect_collections_read(self, ansible_bindings, seeded_distribution, verify_read):
        verify_read(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexApi,
            ansible_bindings.PulpAnsibleApiV3CollectionsApi,
        )

    def test_redirect_collections_versions_read(
        self, ansible_bindings, seeded_distribution, verify_read
    ):
        verify_read(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexVersionsApi,
            ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi,
            version="0.0.3",
        )

    def test_redirect_collections_versions_docs_blob_read(
        self, ansible_bindings, seeded_distribution, verify_read
    ):
        verify_read(
            getattr(
                ansible_bindings,
                "PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexVersionsDocsBlobApi",
            ),
            ansible_bindings.PulpAnsibleApiV3CollectionsVersionsDocsBlobApi,
            version="0.0.3",
        )

    @pytest.mark.skip("https://github.com/OpenAPITools/openapi-generator/issues/20661")
    def test_redirect_collection_versions_all(
        self, ansible_bindings, seeded_distribution, verify_all_collections
    ):
        verify_all_collections(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsAllVersionsApi,
            ansible_bindings.PulpAnsibleApiV3CollectionVersionsAllApi,
        )

    def test_redirect_collections_all(
        self, ansible_bindings, seeded_distribution, verify_all_collections
    ):
        verify_all_collections(
            ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentCollectionsAllCollectionsApi,
            ansible_bindings.PulpAnsibleApiV3CollectionsAllApi,
        )



from gettext import gettext as _

from pulpcore.plugin.exceptions import PulpException


class CollectionNotFound(PulpException):
    error_code = "PLPAN01"

    def __init__(self, namespace, name, url):
        """
        :param url: The full URL that failed validation.
        :type url: str
        """
        # Work around a sudden api change in pulpcore 3.103.
        try:
            super().__init__()
        except BaseException:
            super().__init__(self.error_code)
        self.namespace = namespace
        self.name = name
        self.url = url

    def __str__(self):
        return f"[{self.error_code}] " + _(
            "Collection {namespace}.{name} does not exist on {url}"
        ).format(namespace=self.namespace, name=self.name, url=self.url)



# Generated by Django 4.2.16 on 2024-11-21 08:53

from django.db import migrations
from pulpcore.plugin.migrations import RequireVersion


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0059_collectionversion_unique_sha256"),
    ]

    operations = [
        RequireVersion("ansible", "0.23.0"),
        migrations.RemoveConstraint(
            model_name="collectionversion",
            name="unique_is_highest",
        ),
        migrations.RemoveField(
            model_name="collectionversion",
            name="is_highest",
        ),
    ]



"""
Tests PulpExporter and PulpExport functionality.

NOTE: assumes ALLOWED_EXPORT_PATHS setting contains "/tmp" - all tests will fail if this is not
the case.
"""

import pytest
import uuid

from pulpcore.client.pulp_ansible import AnsibleRepositorySyncURL
from pulpcore.app import settings
from pulp_ansible.tests.functional.constants import ANSIBLE_FIXTURE_URL


# As long as we cannot run this test in domains, it must not run as parallel.
# Orphan cleanup is unsafe in any scenario here.


pytestmark = [
    pytest.mark.skipif(
        "/tmp" not in settings.ALLOWED_EXPORT_PATHS,
        reason="Cannot run export-tests unless /tmp is in ALLOWED_EXPORT_PATHS "
        f"({settings.ALLOWED_EXPORT_PATHS}).",
    ),
]


@pytest.mark.parametrize("cleanup", [True, False])
def test_export_then_import(
    pulpcore_bindings,
    ansible_bindings,
    gen_object_with_cleanup,
    ansible_repo_factory,
    ansible_role_remote_factory,
    build_and_upload_collection,
    ansible_distribution_factory,
    ascii_armored_detached_signing_service,
    random_image_factory,
    monitor_task,
    monitor_task_group,
    cleanup,
):
    """Issue and evaluate a PulpExport (tests both Create and Read)."""
    # Prepare content
    remote_b = ansible_role_remote_factory(url=ANSIBLE_FIXTURE_URL)
    repo_b = ansible_repo_factory(remote=remote_b.pulp_href)
    sync_b_task = ansible_bindings.RepositoriesAnsibleApi.sync(
        repo_b.pulp_href, AnsibleRepositorySyncURL(remote=remote_b.pulp_href)
    ).task
    repo_a = ansible_repo_factory()
    distro = ansible_distribution_factory(repository=repo_a)
    collection, collection_href = build_and_upload_collection(repo_a)
    signing_body = {
        "signing_service": ascii_armored_detached_signing_service.pulp_href,
        "content_units": ["*"],
    }
    signing_task = ansible_bindings.RepositoriesAnsibleApi.sign(repo_a.pulp_href, signing_body).task

    mark_body = {
        "content_units": ["*"],
        "value": "exportable-mark",
    }
    mark_task = ansible_bindings.RepositoriesAnsibleApi.mark(repo_a.pulp_href, mark_body).task

    deprecation_task = ansible_bindings.ContentCollectionDeprecationsApi.create(
        {
            "namespace": collection.namespace,
            "name": collection.name,
            "repository": repo_a.pulp_href,
        }
    ).task

    namespace_body = {
        "name": collection.namespace,
        "avatar": str(random_image_factory()),
        "links": [
            {"name": "homepage", "url": "https://example.com"},
            {"name": "github", "url": "https://gh.com"},
        ],
    }
    namespace_task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.create(
        path=distro.base_path, distro_base_path=distro.base_path, **namespace_body
    ).task

    # Collect all the content create tasks
    monitor_task(sync_b_task)
    monitor_task(signing_task)
    monitor_task(mark_task)
    monitor_task(deprecation_task)
    monitor_task(namespace_task)
    # Gather statistics of the original data
    repo_ver_a = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        f"{repo_a.pulp_href}versions/5/"
    )
    repo_ver_b = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        f"{repo_b.pulp_href}versions/1/"
    )

    # Prepare export
    exporter = gen_object_with_cleanup(
        pulpcore_bindings.ExportersPulpApi,
        {
            "name": str(uuid.uuid4()),
            "path": f"/tmp/{uuid.uuid4()}/",
            "repositories": [repo.pulp_href for repo in [repo_a, repo_b]],
        },
    )

    # Export
    task = pulpcore_bindings.ExportersPulpExportsApi.create(exporter.pulp_href, {}).task
    task = monitor_task(task)
    assert len(task.created_resources) == 1
    export = pulpcore_bindings.ExportersPulpExportsApi.read(task.created_resources[0])
    assert export is not None
    assert len(exporter.repositories) == len(export.exported_resources)
    assert export.output_file_info is not None
    for an_export_filename in export.output_file_info.keys():
        assert "//" not in an_export_filename
    export_filename = next(
        f for f in export.output_file_info.keys() if f.endswith("tar.gz") or f.endswith("tar")
    )

    if cleanup:
        # Remove all content from pulp (currently only the genrated content in repo a).
        # Since repo b is synced we cannot safely do this.
        content_a = pulpcore_bindings.ContentApi.list(
            repository_version=f"{repo_a.pulp_href}versions/5/"
        )
        assert content_a.count == 5
        assert content_a.next is None, "Unexpected pagination. Please redesign the test."
        content_a_hrefs = [c.pulp_href for c in content_a.results]
        monitor_task(ansible_bindings.RepositoriesAnsibleApi.delete(repo_a.pulp_href).task)
        cleanup_result = monitor_task(
            pulpcore_bindings.OrphansCleanupApi.cleanup(
                {"orphan_protection_time": 0, "content_hrefs": content_a_hrefs}
            ).task
        )
        assert (
            next(
                (
                    report
                    for report in cleanup_result.progress_reports
                    if report.code == "clean-up.content"
                )
            ).total
            == 5
        )

    # Prepare import
    repo_c = ansible_repo_factory()
    repo_d = ansible_repo_factory()
    repo_mapping = {repo_a.name: repo_c.name, repo_b.name: repo_d.name}
    importer = gen_object_with_cleanup(
        pulpcore_bindings.ImportersPulpApi,
        {"name": str(uuid.uuid4()), "repo_mapping": repo_mapping},
    )

    # Import
    import_response = pulpcore_bindings.ImportersPulpImportsApi.create(
        importer.pulp_href, {"path": export_filename}
    )
    monitor_task_group(import_response.task_group)
    repo_c = ansible_bindings.RepositoriesAnsibleApi.read(repo_c.pulp_href)
    repo_d = ansible_bindings.RepositoriesAnsibleApi.read(repo_d.pulp_href)
    assert repo_c.latest_version_href == f"{repo_c.pulp_href}versions/1/"
    assert repo_d.latest_version_href == f"{repo_d.pulp_href}versions/1/"
    repo_ver_c = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        f"{repo_c.pulp_href}versions/1/"
    )
    repo_ver_d = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        f"{repo_d.pulp_href}versions/1/"
    )
    assert (
        repo_ver_c.content_summary.added["ansible.collection_version"]["count"]
        == repo_ver_a.content_summary.present["ansible.collection_version"]["count"]
    )
    assert (
        repo_ver_c.content_summary.added["ansible.collection_deprecation"]["count"]
        == repo_ver_a.content_summary.present["ansible.collection_deprecation"]["count"]
    )
    assert (
        repo_ver_c.content_summary.added["ansible.collection_signature"]["count"]
        == repo_ver_a.content_summary.present["ansible.collection_signature"]["count"]
    )
    assert (
        repo_ver_c.content_summary.added["ansible.collection_mark"]["count"]
        == repo_ver_a.content_summary.present["ansible.collection_mark"]["count"]
    )
    assert (
        repo_ver_c.content_summary.added["ansible.namespace"]["count"]
        == repo_ver_a.content_summary.present["ansible.namespace"]["count"]
    )
    assert (
        repo_ver_d.content_summary.added["ansible.role"]["count"]
        == repo_ver_b.content_summary.present["ansible.role"]["count"]
    )

    # Import a second time
    import_response = pulpcore_bindings.ImportersPulpImportsApi.create(
        importer.pulp_href, {"path": export_filename}
    )
    monitor_task_group(import_response.task_group)
    assert len(pulpcore_bindings.ImportersPulpImportsApi.list(importer.pulp_href).results) == 2
    for repo in [repo_c, repo_d]:
        repo = ansible_bindings.RepositoriesAnsibleApi.read(repo.pulp_href)
        # still only one version as pulp won't create a new version if nothing changed
        assert repo.latest_version_href == f"{repo.pulp_href}versions/1/"



# Generated by Django 4.2.16 on 2024-12-05 12:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0037_gitremote'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionVersionSignature',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionversionsignature', serialize=False, to='core.content')),
                ('data', models.BinaryField()),
                ('digest', models.CharField(max_length=64)),
                ('pubkey_fingerprint', models.CharField(max_length=64)),
                ('signed_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='signatures', to='ansible.collectionversion')),
                ('signing_service', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='signatures', to='core.signingservice')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('pubkey_fingerprint', 'signed_collection')},
            },
            bases=('core.content',),
        ),
    ]



import pytest

from pulpcore.client.pulp_ansible.exceptions import ApiException

from pulp_ansible.tests.functional.utils import content_counts, randstr


def test_collection_deletion(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Test deleting an entire collection."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(path=distribution.base_path)
    assert collections.meta.count == 1

    task = monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        ).task
    )
    assert len(task.created_resources) == 1

    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(distribution.base_path)
    assert collections.meta.count == 0

    versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        path=distribution.base_path,
        name=collection_name,
        namespace=collection_namespace,
    )

    assert versions.meta.count == 0

    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.read(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        )

    assert exc_info.value.status == 404


def test_collection_version_deletion(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Test deleting a specific collection version."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(distribution.base_path)
    assert collections.meta.count == 1

    versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        path=distribution.base_path,
        name=collection_name,
        namespace=collection_namespace,
    )
    assert versions.meta.count == 2

    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.0",
        ).task
    )

    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.read(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.0",
        )
    assert exc_info.value.status == 404

    # Verify that the collection still exists
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(distribution.base_path)
    assert collections.meta.count == 1

    # Verify that the other versions still exist
    versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        path=distribution.base_path,
        name=collection_name,
        namespace=collection_namespace,
    )
    assert versions.meta.count == 1

    # Delete the other versions

    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.1",
        ).task
    )

    # Verify all the versions have been deleted
    versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        path=distribution.base_path,
        name=collection_name,
        namespace=collection_namespace,
    )
    assert versions.meta.count == 0

    # With all the versions deleted, verify that the collection has also
    # been deleted
    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.read(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        )
    assert exc_info.value.status == 404


def test_invalid_deletion(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Test deleting collections that are dependencies for other collections."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    dependent_collection, dependent_collection_href = build_and_upload_collection(
        repository,
        config={"dependencies": {f"{collection_namespace}.{collection_name}": "1.0.0"}},
    )

    err_msg = f"{dependent_collection.namespace}.{dependent_collection.name} 1.0.0"

    # Verify entire collection can't be deleted
    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        )

    # check error message includes collection that's blocking delete
    assert err_msg in exc_info.value.body

    # Verify specific version that's used can't be deleted
    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.0",
        )
    assert exc_info.value.status == 400

    # check error message includes collection that's blocking delete
    assert err_msg in exc_info.value.body

    # Verify non dependent version can be deleted.
    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.1",
        ).task
    )


def test_delete_deprecated_content(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Test that deprecated content is removed correctly."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    # Deprecate the collection
    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.update(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            patched_collection={"deprecated": True},
        ).task
    )

    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        ).task
    )

    # Verify that all the content is gone
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    latest_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )

    assert content_counts(latest_version) == {}


def test_delete_signed_content(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    ascii_armored_detached_signing_service,
    monitor_task,
):
    """Test that signature content is removed correctly."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    # Sign the collections
    body = {
        "content_units": [
            "*",
        ],
        "signing_service": ascii_armored_detached_signing_service.pulp_href,
    }
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sign(repository.pulp_href, body).task)

    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
        ).task
    )

    # Verify that all the content is gone
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    latest_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )

    assert content_counts(latest_version) == {}


def test_version_deletion_with_range_of_versions(
    ansible_bindings,
    ansible_distribution_factory,
    ansible_repository_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Verify collections can be deleted when another version satisfies requirements."""
    repository = ansible_repository_factory()
    distribution = ansible_distribution_factory(repository=repository)

    collection_name = randstr()
    collection_namespace = randstr()
    for version in ["1.0.0", "1.0.1"]:
        build_and_upload_collection(
            repository,
            config={"name": collection_name, "namespace": collection_namespace, "version": version},
        )

    # Create a collection that depends on any version of an existing collection
    dependent_collection, dependent_collection_href = build_and_upload_collection(
        repository,
        config={"dependencies": {f"{collection_namespace}.{collection_name}": "*"}},
    )

    # Verify the collection version can be deleted as long as there is one version
    # left that satisfies the requirements.
    monitor_task(
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.0",
        ).task
    )

    # Verify that the last version of the collection can't be deleted

    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.delete(
            path=distribution.base_path,
            name=collection_name,
            namespace=collection_namespace,
            version="1.0.1",
        )

    assert exc_info.value.status == 400



# Generated by Django 3.2.14 on 2022-07-15 22:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0055_alter_collectionversion_version_alter_role_version'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionversion',
            name='sha256',
            field=models.CharField(default='', max_length=64, null=True),
            preserve_default=False,
        ),
    ]
